#ifndef _HANDLE_DATA_HPP_
#define _HANDLE_DATA_HPP_

#include <chronoptics/tof.h>

#include <functional>
#include <iostream>
#include <mex.hpp>
#include <mexAdapter.hpp>
#include <sstream>

namespace impl {

constexpr size_t user_pointer_size = 100;

using namespace matlab;

inline void error_(std::string message) {
  throw std::runtime_error(std::move(message));
}

inline void check_error_(tof_error_t error) {
  if (!error) return;

  std::string message(tof_error_message(error));
  tof_error_delete(error);

  error_(message);
}

inline std::vector<tof_frame_type> get_frame_types(tof_camera_t camera) {
  tof_error_t error{nullptr};

  // Retrieve set stream list
  size_t capacity =
      tof_camera_get_output_frame_types(camera, nullptr, 0, &error);
  check_error_(error);

  std::vector<tof_frame_type> types(capacity);
  capacity =
      tof_camera_get_output_frame_types(camera, types.data(), capacity, &error);
  check_error_(error);

  types.resize(capacity);

  return types;
}

inline size_t get_maximum_frame_size(tof_camera_t camera) {
  tof_error_t error{nullptr};

  const auto config = tof_camera_get_camera_config(camera, &error);
  check_error_(error);

  const auto roi = tof_camera_config_get_roi(config, 0, &error);
  check_error_(error);

  const int32_t rows = tof_roi_sensor_rows(roi, &error);
  check_error_(error);

  const int32_t cols = tof_roi_sensor_cols(roi, &error);
  check_error_(error);

  tof_camera_config_delete(config);
  tof_roi_delete(roi);

  return rows * cols;
}

inline size_t get_rgb_frame_size(tof_camera_t camera) {
  tof_error_t error{nullptr};

  const auto config = tof_camera_get_camera_config(camera, &error);
  check_error_(error);

  size_t rows = tof_camera_config_get_rgb_height(config, &error);
  if (error) return 0;

  size_t cols = tof_camera_config_get_rgb_width(config, &error);
  if (error) return 0;

  return rows * cols;
}

class Released {
 public:
  Released(tof_frame_type frame_type, size_t frame_size)
      : frame_type_(frame_type), frame_size_(frame_size) {}

  bool is_released() { return released_; }
  void release() { released_ = true; }

  tof_frame_type frame_type() { return frame_type_; }
  size_t frame_size() { return frame_size_; }

 protected:
 private:
  tof_frame_type frame_type_;
  size_t frame_size_;
  bool released_{false};
};

template <typename T>
class UserPointer : public Released {
 public:
  UserPointer(tof_frame_type frame_type, data::ArrayFactory& factory,
              size_t size)
      : Released(frame_type, size), buffer_(factory.createBuffer<T>(size)) {}

  data::buffer_ptr_t<T> buffer() { return std::move(buffer_); }

 protected:
  data::buffer_ptr_t<T> buffer_;
};

template <typename T>
class CameraHandler : public UserPointer<T> {
 public:
  CameraHandler(tof_camera_t camera, tof_frame_type frame_type,
                data::ArrayFactory& factory, size_t size)
      : UserPointer<T>(frame_type, factory, size), camera_(camera) {
    add_user_pointer();
  }

  tof_user_pointer_destructed_t cb = [](uint8_t*, size_t, void* user_data) {
    auto data = static_cast<CameraHandler<T>*>(user_data);
    if (data->is_released()) {
      delete data;
      return;
    }

    data->add_user_pointer();
  };

 private:
  void add_user_pointer() {
    tof_error_t error{nullptr};
    tof_camera_add_user_pointer(
        camera_, reinterpret_cast<uint8_t*>(this->buffer_.get()),
        this->frame_size() * sizeof(T), cb, this, this->frame_type(), &error);

    if (error) {
      // Failed to add user pointer, clean up
      tof_error_delete(error);
      delete this;
    }
  }

  tof_camera_t camera_;
};

inline void queue_user_pointer(tof_camera_t camera, tof_frame_type frame_type,
                               size_t frame_size, data::ArrayFactory& factory,
                               bool first_queue) {
  switch (frame_type) {
    case TOF_FRAME_TYPE_RAW_COMMON:
    case TOF_FRAME_TYPE_COMMON:
      new CameraHandler<uint16_t>(camera, frame_type, factory, frame_size);
      break;
    case TOF_FRAME_TYPE_RAW:
      new CameraHandler<int16_t>(camera, frame_type, factory, frame_size);
      break;
    case TOF_FRAME_TYPE_RAW_AB:
      new CameraHandler<uint16_t>(camera, frame_type, factory,
                                  (first_queue ? frame_size * 2 : frame_size));
      break;
    case TOF_FRAME_TYPE_REAL:
    case TOF_FRAME_TYPE_IMAGINARY:
    case TOF_FRAME_TYPE_PHASE:
    case TOF_FRAME_TYPE_AMPLITUDE:
    case TOF_FRAME_TYPE_X:
    case TOF_FRAME_TYPE_Y:
    case TOF_FRAME_TYPE_Z:
      new CameraHandler<float>(camera, frame_type, factory, frame_size);
      break;
    case TOF_FRAME_TYPE_INTENSITY:
      new CameraHandler<uint8_t>(camera, frame_type, factory, frame_size);
      break;
    case TOF_FRAME_TYPE_RADIAL:
      new CameraHandler<uint16_t>(camera, frame_type, factory, frame_size);
      break;
    case TOF_FRAME_TYPE_BGR_PROJECTED:
    case TOF_FRAME_TYPE_BGR:
      new CameraHandler<uint8_t>(camera, frame_type, factory,
                                 (first_queue ? frame_size * 3 : frame_size));
      break;
    default: {
      std::ostringstream stream;
      stream << "Frame type " << tof_frame_type_to_string(frame_type, nullptr)
             << " is not supported in matlab";
      error_(stream.str());
      break;
    }
  }
}

inline void tofCameraStart(mex::ArgumentList outputs, mex::ArgumentList inputs,
                           data::ArrayFactory& factory) {
  if (inputs.size() != 2 || outputs.size() != 0) {
    error_("2 inputs are required and 0 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be Camera");
  }

  auto camera =
      reinterpret_cast<tof_camera_t>(static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  size_t frame_size = get_maximum_frame_size(camera);
  size_t rgb_frame_size = get_rgb_frame_size(camera);

  auto frame_types = get_frame_types(camera);

  tof_camera_set_user_pointer_capacity(camera, user_pointer_size, &error);
  check_error_(error);

  for (auto frame_type : frame_types) {
    auto size =
        (frame_type == TOF_FRAME_TYPE_BGR ? rgb_frame_size : frame_size);

    for (size_t i = 0; i < user_pointer_size; i++) {
      queue_user_pointer(camera, frame_type, size, factory, true);
    }
  }

  // Actually start the camera
  tof_camera_start(camera, &error);
  check_error_(error);
}

inline void tofCameraGetFrames(mex::ArgumentList outputs,
                               mex::ArgumentList inputs,
                               data::ArrayFactory& factory) {
  if (inputs.size() != 2 || outputs.size() != 1) {
    error_("2 inputs are required and 1 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be Camera");
  }

  auto camera =
      reinterpret_cast<tof_camera_t>(static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  size_t capacity = tof_camera_get_frames(camera, nullptr, 0, &error);
  check_error_(error);
  std::vector<tof_data_t> frames(capacity);
  capacity = tof_camera_get_frames(camera, frames.data(), capacity, &error);
  check_error_(error);
  std::vector<uint64_t> output;
  output.reserve(capacity);
  for (auto item : frames) {
    output.push_back(reinterpret_cast<uint64_t>(item));
  }

  for (auto data : frames) {
    bool user_allocated = tof_data_user_allocated(data, &error);
    check_error_(error);

    if (!user_allocated) continue;

    auto user_pointer = tof_data_user_pointer(data, &error);
    check_error_(error);

    auto base = static_cast<Released*>(user_pointer);
    base->release();

    queue_user_pointer(camera, base->frame_type(), base->frame_size(), factory,
                       false);
  }

  outputs[0] = factory.createArray({1, capacity}, output.data(),
                                   output.data() + capacity);
}

template <typename T>
void create_output(mex::ArgumentList outputs, void* handler_ptr,
                   data::ArrayDimensions& dims, data::ArrayFactory& factory) {
  auto ptr = static_cast<UserPointer<T>*>(handler_ptr);

  auto buffer = ptr->buffer();

  if (buffer.get() == nullptr) {
    throw std::runtime_error("Can only call data once");
  }

  outputs[0] = factory.createArrayFromBuffer(dims, std::move(buffer));
}

inline void set_output(mex::ArgumentList outputs, void* handler_ptr,
                       tof_frame_type frame_type, data::ArrayDimensions& dims,
                       data::ArrayFactory& factory) {
  switch (frame_type) {
    case TOF_FRAME_TYPE_RAW_COMMON:
    case TOF_FRAME_TYPE_COMMON:
      return create_output<uint16_t>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_RAW:
      return create_output<int16_t>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_RAW_AB:
      // dims[0] *= 2;
      return create_output<uint16_t>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_REAL:
    case TOF_FRAME_TYPE_IMAGINARY:
    case TOF_FRAME_TYPE_PHASE:
    case TOF_FRAME_TYPE_AMPLITUDE:
    case TOF_FRAME_TYPE_X:
    case TOF_FRAME_TYPE_Y:
    case TOF_FRAME_TYPE_Z:
      return create_output<float>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_INTENSITY:
      return create_output<uint8_t>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_RADIAL:
      return create_output<uint16_t>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_BGR_PROJECTED:
    case TOF_FRAME_TYPE_BGR:
      dims.insert(dims.begin(), 3);
      return create_output<uint8_t>(outputs, handler_ptr, dims, factory);
    default: {
      std::ostringstream stream;
      stream << "Frame type " << tof_frame_type_to_string(frame_type, nullptr)
             << " is not supported in matlab";
      error_(stream.str());
      break;
    }
  }
}

inline void tofDataGetData(mex::ArgumentList outputs, mex::ArgumentList inputs,
                           data::ArrayFactory& factory) {
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be Data");
  }

  auto data = reinterpret_cast<tof_data_t>(static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  bool user_allocated = tof_data_user_allocated(data, &error);
  check_error_(error);

  if (!user_allocated) {
    error_("ToF data is not user allocated");
  }

  auto handler_ptr = tof_data_user_pointer(data, &error);
  check_error_(error);

  size_t rows = tof_data_rows(data, &error);
  check_error_(error);

  size_t cols = tof_data_cols(data, &error);
  check_error_(error);

  auto frame_type = tof_data_frame_type(data, &error);
  check_error_(error);

  data::ArrayDimensions dims{cols, rows};
  set_output(outputs, handler_ptr, frame_type, dims, factory);
}

template <typename T>
inline void copy_data(mex::ArgumentList inputs, void* data_ptr, size_t rows,
                      size_t cols, data::ArrayType type) {
  const size_t size = rows * cols;

  if (inputs[2].getType() != type) {
    error_("Input matrix type does not match frame type");
  }

  if (inputs[2].getNumberOfElements() != size ||
      inputs[2].getDimensions()[0] != cols) {
    error_("Input matrix size does not match");
  }

  const matlab::data::TypedArray<T> arr_t = inputs[2];
  matlab::data::TypedIterator<const T> it(arr_t.begin());
  const T* out_ptr = it.operator->();
  T* in_ptr = reinterpret_cast<T*>(data_ptr);

  std::copy(out_ptr, out_ptr + size, in_ptr);
}

inline void copy_data_select(mex::ArgumentList inputs, void* data_ptr,
                             size_t rows, size_t cols,
                             tof_frame_type frame_type) {
  switch (frame_type) {
    case TOF_FRAME_TYPE_RAW_COMMON:
    case TOF_FRAME_TYPE_COMMON:
    case TOF_FRAME_TYPE_RAW_AB:
    case TOF_FRAME_TYPE_RADIAL:
      return copy_data<uint16_t>(inputs, data_ptr, rows, cols,
                                 data::ArrayType::UINT16);
    case TOF_FRAME_TYPE_RAW:
      return copy_data<int16_t>(inputs, data_ptr, rows, cols,
                                data::ArrayType::INT16);
    case TOF_FRAME_TYPE_REAL:
    case TOF_FRAME_TYPE_IMAGINARY:
    case TOF_FRAME_TYPE_PHASE:
    case TOF_FRAME_TYPE_AMPLITUDE:
    case TOF_FRAME_TYPE_X:
    case TOF_FRAME_TYPE_Y:
    case TOF_FRAME_TYPE_Z:
      return copy_data<float>(inputs, data_ptr, rows, cols,
                              data::ArrayType::COMPLEX_SINGLE);
    case TOF_FRAME_TYPE_INTENSITY:
      return copy_data<uint8_t>(inputs, data_ptr, rows, cols,
                                data::ArrayType::UINT8);
    default: {
      std::ostringstream stream;
      stream << "Frame type " << tof_frame_type_to_string(frame_type, nullptr)
             << " is not supported in setData";
      error_(stream.str());
      break;
    }
  }
}

inline void tofDataSetData(mex::ArgumentList inputs) {
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be Data");
  }

  auto data = reinterpret_cast<tof_data_t>(static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  auto ptr = tof_data_data(data, &error);
  check_error_(error);

  size_t rows = tof_data_rows(data, &error);
  check_error_(error);

  size_t cols = tof_data_cols(data, &error);
  check_error_(error);

  auto frame_type = tof_data_frame_type(data, &error);
  check_error_(error);

  copy_data_select(inputs, ptr, rows, cols, frame_type);
}

inline void tofDataData(mex::ArgumentList outputs, mex::ArgumentList inputs,
                        data::ArrayFactory& factory) {
  bool correct_get = inputs.size() == 2 && outputs.size() == 1;
  bool correct_set = inputs.size() == 3 && outputs.size() == 0;

  if (correct_get) {
    tofDataGetData(outputs, inputs, factory);
  } else if (correct_set) {
    tofDataSetData(inputs);
  } else {
    error_("wrong inputs/outputs supplied expecting either 2/1 or 3/0");
  }
}

}  // namespace impl

#endif
