import ctypes
from typing import List, Optional
from .__library__ import lib, get_error
from .stream import *
from .processing_config import *
from .data import *
from .camera_config import *
from .calibration import *


class Camera():
    """
    The main interface to the depth cameras
    """
    def __del__(self):
        if self._ptr:
            lib.tof_camera_delete(self._ptr)
            self._ptr = ctypes.c_void_p()


    def __init__(self, ptr: Optional[ctypes.c_void_p] = None):
        """
        Construct new camera class
        """
        self._ptr = ctypes.c_void_p()

        if ptr != None:
            self._ptr = ptr
        else:
            raise RuntimeError("Can't construct empty class")

    def start(self) -> None:
        """
        Start streaming of the camera
        """
        return lib.tof_camera_start(self._ptr, get_error())


    def stop(self) -> None:
        """
        Stop streaming of the camera
        """
        return lib.tof_camera_stop(self._ptr, get_error())


    def isStreaming(self) -> bool:
        """
        Check whether the camera is streaming

        Returns
        ----------
        bool
            Is streaming
        """
        return lib.tof_camera_is_streaming(self._ptr, get_error())


    def isConnected(self) -> bool:
        """
        Check whether the camera is still connected

        Returns
        ----------
        bool
            Is connected
        """
        return lib.tof_camera_is_connected(self._ptr, get_error())


    def getStreamList(self) -> List[Stream]:
        """
        Get the different output streams the camera can provide

        Returns
        ----------
        List[Stream]
            The streams
        """
        size = lib.tof_camera_get_stream_list(self._ptr, ctypes.c_void_p(), 0, get_error())
        array = (ctypes.c_void_p * size)()
        lib.tof_camera_get_stream_list(self._ptr, array, size, get_error())
        return [Stream(ptr = item) for item in array]

    def getProcessConfig(self) -> ProcessingConfig:
        """
        the processing config of the camera

        Returns
        ----------
        ProcessingConfig
            The processing config
        """
        return ProcessingConfig(ptr = lib.tof_camera_get_process_config(self._ptr, get_error()))


    def hasFrames(self) -> bool:
        """
        Check if the camera has frames available

        Returns
        ----------
        bool
            Has frames available
        """
        return lib.tof_camera_has_frames(self._ptr, get_error())


    def getFrames(self) -> List[Data]:
        """
        Get data frames from camera

        Returns
        ----------
        List[Data]
            Data frames
        """
        size = lib.tof_camera_get_frames(self._ptr, ctypes.c_void_p(), 0, get_error())
        array = (ctypes.c_void_p * size)()
        lib.tof_camera_get_frames(self._ptr, array, size, get_error())
        return [Data(ptr = item) for item in array]

    def clearBuffer(self) -> None:
        """
        Clear the circular buffer used to store frames
        """
        return lib.tof_camera_clear_buffer(self._ptr, get_error())


    def getBufferSize(self) -> int:
        """
        Get the circular buffer size

        Returns
        ----------
        int
            The maximum number of sets of frames that is stored before the oldest is
            overwritten
        """
        return lib.tof_camera_get_buffer_size(self._ptr, get_error())


    def setBufferSize(self, size: int) -> None:
        """
        Set the circular buffer size

        Parameters
        ----------
        size : int
            The maximum number of sets of frames that is stored before the oldest is
            overwritten
        """
        return lib.tof_camera_set_buffer_size(self._ptr, size, get_error())


    def getCameraConfig(self) -> CameraConfig:
        """
        Get the currently active camera config

        Returns
        ----------
        CameraConfig
            The camera config
        """
        return CameraConfig(ptr = lib.tof_camera_get_camera_config(self._ptr, get_error()))


    def getSerial(self) -> str:
        """
        Get the serial number of the current camera

        Returns
        ----------
        str
            The serial
        """
        return lib.tof_camera_get_serial(self._ptr, get_error()).decode("utf-8")


    def getCalibration(self) -> Calibration:
        """
        Get calibration from the camera

        Returns
        ----------
        Calibration
            The calibration
        """
        return Calibration(ptr = lib.tof_camera_get_calibration(self._ptr, get_error()))


    def getUserPointerCapacity(self) -> int:
        """
        Get the amount of user pointers that can be stored

        Returns
        ----------
        int
            The capacity
        """
        return lib.tof_camera_get_user_pointer_capacity(self._ptr, get_error())


    def setUserPointerCapacity(self, capacity: int) -> None:
        """
        Set the amount of user pointers that can be stored

        Parameters
        ----------
        capacity : int
            The capacity
        """
        return lib.tof_camera_set_user_pointer_capacity(self._ptr, capacity, get_error())


    def clearUserPointers(self) -> None:
        """
        Clear all user pointers in memory
        """
        return lib.tof_camera_clear_user_pointers(self._ptr, get_error())


    def configIndex(self) -> int:
        """
        Get the currently set configuration index

        Returns
        ----------
        int
            config index
        """
        return lib.tof_camera_config_index(self._ptr, get_error())


    def switchConfig(self, config_index: int) -> None:
        """
        Switch to a different config in real time, depending on configuration and
        camera, this can take 100-300ms

        Parameters
        ----------
        config_index : int
            Config index to switch to
        """
        return lib.tof_camera_switch_config(self._ptr, config_index, get_error())


    def possibleFrameTypes(self) -> List[FrameType]:
        """
        Get the frame types that this camera can output

        Returns
        ----------
        List[FrameType]
            Possible frame types
        """
        size = lib.tof_camera_possible_frame_types(self._ptr, ctypes.c_int(), 0, get_error())
        array = (ctypes.c_int * size)()
        lib.tof_camera_possible_frame_types(self._ptr, array, size, get_error())
        return array[:]


    def outputFrameTypesSet(self) -> bool:
        """
        Whether output frame types have been set

        Returns
        ----------
        bool
            Output frame types set
        """
        return lib.tof_camera_output_frame_types_set(self._ptr, get_error())


    def getOutputFrameTypes(self) -> List[FrameType]:
        """
        Get the frame types the camera should output

        Returns
        ----------
        List[FrameType]
            Output frame types
        """
        size = lib.tof_camera_get_output_frame_types(self._ptr, ctypes.c_int(), 0, get_error())
        array = (ctypes.c_int * size)()
        lib.tof_camera_get_output_frame_types(self._ptr, array, size, get_error())
        return array[:]


    def setOutputFrameTypes(self, output_frame_types: List[FrameType]) -> None:
        """
        Set the frame types the camera should output

        Parameters
        ----------
        output_frame_types : List[FrameType]
            Output frame types
        """
        return lib.tof_camera_set_output_frame_types(self._ptr, (ctypes.c_int * len(output_frame_types))(*output_frame_types), len(output_frame_types), get_error())


    def hasMultipleConfigurations(self) -> bool:
        """
        Check if multiple configurations are set. Switch config can only be used
        when this is the case.

        Returns
        ----------
        bool
            Is set
        """
        return lib.tof_camera_has_multiple_configurations(self._ptr, get_error())


    def getMultipleCameraConfigurations(self) -> List[CameraConfig]:
        """
        Get all the set camera configurations

        Returns
        ----------
        List[CameraConfig]
            Multiple camera configurations
        """
        size = lib.tof_camera_get_multiple_camera_configurations(self._ptr, ctypes.c_void_p(), 0, get_error())
        array = (ctypes.c_void_p * size)()
        lib.tof_camera_get_multiple_camera_configurations(self._ptr, array, size, get_error())
        return [CameraConfig(ptr = item) for item in array]

    def getMultipleProcessingConfigurations(self) -> List[ProcessingConfig]:
        """
        Get all the set processing configurations

        Returns
        ----------
        List[ProcessingConfig]
            Multiple processing configurations
        """
        size = lib.tof_camera_get_multiple_processing_configurations(self._ptr, ctypes.c_void_p(), 0, get_error())
        array = (ctypes.c_void_p * size)()
        lib.tof_camera_get_multiple_processing_configurations(self._ptr, array, size, get_error())
        return [ProcessingConfig(ptr = item) for item in array]

    def getMultipleNames(self) -> List[str]:
        """
        Get the name for each configuration

        Returns
        ----------
        List[str]
            Names
        """
        size = lib.tof_camera_get_multiple_names(self._ptr, ctypes.c_char_p(), 0, get_error())
        array = (ctypes.c_char_p * size)()
        lib.tof_camera_get_multiple_names(self._ptr, array, size, get_error())
        return array[:]


    def getMultipleDescriptions(self) -> List[str]:
        """
        Get the description for each configuration

        Returns
        ----------
        List[str]
            Descriptions
        """
        size = lib.tof_camera_get_multiple_descriptions(self._ptr, ctypes.c_char_p(), 0, get_error())
        array = (ctypes.c_char_p * size)()
        lib.tof_camera_get_multiple_descriptions(self._ptr, array, size, get_error())
        return array[:]


class PersistentIp():
    """
    A persistent ip data structure
    """
    def __del__(self):
        if self._ptr:
            lib.tof_persistent_ip_delete(self._ptr)
            self._ptr = ctypes.c_void_p()


    def __init__(self, ip: Optional[str] = None, netmask: Optional[str] = None, gateway: Optional[str] = None, enabled: Optional[bool] = None, ptr: Optional[ctypes.c_void_p] = None):
        """
        Construct new persistent_ip class
        ----------

        Create persistent ip

        Parameters
        ----------
        ip : str
                IP address
        netmask : str
                Netmask
        gateway : str
                Gateway
        enabled : bool
                Enable or disable persistent ip
        """
        self._ptr = ctypes.c_void_p()

        if ptr != None:
            self._ptr = ptr
        elif ip != None and netmask != None and gateway != None and enabled != None:
            self._ptr = lib.tof_persistent_ip_new(ip.encode("utf-8"), netmask.encode("utf-8"), gateway.encode("utf-8"), enabled, get_error())
        else:
            raise RuntimeError("Can't construct empty class")

    def getIp(self) -> str:
        """
        Get the IP address

        Returns
        ----------
        str
            the IP address
        """
        return lib.tof_persistent_ip_get_ip(self._ptr, get_error()).decode("utf-8")


    def setIp(self, ip: str) -> None:
        """
        Set the IP address

        Parameters
        ----------
        ip : str
            the IP address
        """
        return lib.tof_persistent_ip_set_ip(self._ptr, ip.encode("utf-8"), get_error())


    def getNetmask(self) -> str:
        """
        Get the netmask

        Returns
        ----------
        str
            The netmask
        """
        return lib.tof_persistent_ip_get_netmask(self._ptr, get_error()).decode("utf-8")


    def setNetmask(self, netmask: str) -> None:
        """
        Set the netmask

        Parameters
        ----------
        netmask : str
            The netmask
        """
        return lib.tof_persistent_ip_set_netmask(self._ptr, netmask.encode("utf-8"), get_error())


    def getGateway(self) -> str:
        """
        Get the gateway

        Returns
        ----------
        str
            The gateway
        """
        return lib.tof_persistent_ip_get_gateway(self._ptr, get_error()).decode("utf-8")


    def setGateway(self, gateway: str) -> None:
        """
        Set the gateway

        Parameters
        ----------
        gateway : str
            The gateway
        """
        return lib.tof_persistent_ip_set_gateway(self._ptr, gateway.encode("utf-8"), get_error())


    def getEnabled(self) -> bool:
        """
        Get if persistent ip is enabled

        Returns
        ----------
        bool
            Is enabled
        """
        return lib.tof_persistent_ip_get_enabled(self._ptr, get_error())


    def setEnabled(self, enabled: bool) -> None:
        """
        Set if persistent ip is enabled

        Parameters
        ----------
        enabled : bool
            Is enabled
        """
        return lib.tof_persistent_ip_set_enabled(self._ptr, enabled, get_error())


class LiveCamera(Camera):
    """
    Functions shared across all live cameras
    """
    def __del__(self):
        if self._ptr:
            lib.tof_live_camera_delete(self._ptr)
            self._ptr = ctypes.c_void_p()


    def __init__(self, ptr: Optional[ctypes.c_void_p] = None):
        """
        Construct new live_camera class
        """
        self._ptr = ctypes.c_void_p()

        if ptr != None:
            self._ptr = ptr
        else:
            raise RuntimeError("Can't construct empty class")

    def version(self) -> str:
        """
        Get the tof library version running on the camera

        Returns
        ----------
        str
            Version
        """
        return lib.tof_live_camera_version(self._ptr, get_error()).decode("utf-8")


    def remoteUpdateCapable(self) -> bool:
        """
        Check whether the camera is capable of updating remotely

        Returns
        ----------
        bool
            Remote update capable
        """
        return lib.tof_live_camera_remote_update_capable(self._ptr, get_error())


    def remoteUpdate(self, file_location: str) -> None:
        """
        Update the camera remotely. After uploading the update file the function
        will return. If a correct update file was uploaded, the camera will flash
        the update and reboot. This generally takes about a minute. The camera can
        also be updated via a web interface by visiting the ip address of the camera
        on port 8080, for example http://192.168.1.208:8080

        Parameters
        ----------
        file_location : str
            Location of the update file
        """
        return lib.tof_live_camera_remote_update(self._ptr, file_location.encode("utf-8"), get_error())


    def persistentIpCapable(self) -> bool:
        """
        Whether the camera is capable of setting persistent ip

        Returns
        ----------
        bool
            Persistent ip capable
        """
        return lib.tof_live_camera_persistent_ip_capable(self._ptr, get_error())


    def getPersistentIp(self) -> PersistentIp:
        """
        Get the current persistent ip settings

        Returns
        ----------
        PersistentIp
            Persistent IP
        """
        return PersistentIp(ptr = lib.tof_live_camera_get_persistent_ip(self._ptr, get_error()))


    def setPersistentIp(self, persistent_ip: PersistentIp) -> None:
        """
        Set the persistent ip settings. To activate persistent ip settings run
        apply_persistent_ip. 

        Parameters
        ----------
        persistent_ip : PersistentIp
            The persistent ip
        """
        return lib.tof_live_camera_set_persistent_ip(self._ptr, persistent_ip._ptr, get_error())


    def applyPersistentIpCapable(self) -> bool:
        """
        Whether the camera is capable of applying the persistent ip

        Returns
        ----------
        bool
            Capable
        """
        return lib.tof_live_camera_apply_persistent_ip_capable(self._ptr, get_error())


    def applyPersistentIp(self) -> None:
        """
        Apply persistent ip by rebooting the camera, power cycling the camera is
        discouraged because the applied ip settings might not stick.
        """
        return lib.tof_live_camera_apply_persistent_ip(self._ptr, get_error())


    def softwareTriggerCapable(self) -> bool:
        """
        Check whether the camera is capable of software trigger

        Returns
        ----------
        bool
            Software trigger capable
        """
        return lib.tof_live_camera_software_trigger_capable(self._ptr, get_error())


    def softwareTrigger(self) -> None:
        """
        Software trigger the camera
        """
        return lib.tof_live_camera_software_trigger(self._ptr, get_error())


    def getDelay(self) -> int:
        """
        Get delay between network packets

        Returns
        ----------
        int
            The delay, depending on the camera software it is in nanoseconds or in
            cpu instruction
        """
        return lib.tof_live_camera_get_delay(self._ptr, get_error())


    def setDelay(self, delay: int) -> None:
        """
        Set delay between network packets

        Parameters
        ----------
        delay : int
            The delay, depending on the camera software it is in nanoseconds or in
            cpu instruction
        """
        return lib.tof_live_camera_set_delay(self._ptr, delay, get_error())


    def getPacketSize(self) -> int:
        """
        Get maximum size of each network packet transmitted

        Returns
        ----------
        int
            Packet size in bytes
        """
        return lib.tof_live_camera_get_packet_size(self._ptr, get_error())


    def setPacketSize(self, size: int) -> None:
        """
        Set maximum size of each network packet transmitted

        Parameters
        ----------
        size : int
            Packet size in bytes
        """
        return lib.tof_live_camera_set_packet_size(self._ptr, size, get_error())


    def getTemperaturesCapable(self) -> bool:
        """
        Whether the camera is capable of returning temperatures

        Returns
        ----------
        bool
            Capable
        """
        return lib.tof_live_camera_get_temperatures_capable(self._ptr, get_error())


    def getTemperatures(self) -> List[float]:
        """
        Get the current on camera temperatures

        Returns
        ----------
        List[float]
            Temperatures
        """
        size = lib.tof_live_camera_get_temperatures(self._ptr, ctypes.c_float(), 0, get_error())
        array = (ctypes.c_float * size)()
        lib.tof_live_camera_get_temperatures(self._ptr, array, size, get_error())
        return array[:]


    def getGpio(self, pin: int) -> bool:
        """
        Get voltage level on given pin

        Parameters
        ----------
        pin : int
            Pin

        Returns
        ----------
        bool
            Voltage 3.3V
        """
        return lib.tof_live_camera_get_gpio(self._ptr, pin, get_error())


    def setGpio(self, pin: int, level: bool) -> None:
        """
        Set voltage level on given pin

        Parameters
        ----------
        pin : int
            Pin
        level : bool
            Voltage level
        """
        return lib.tof_live_camera_set_gpio(self._ptr, pin, level, get_error())


